# -*- coding: utf-8 -*-
# Guix Import Debian --- Import helper for GNU Guix
# Copyright © 2016 Hartmut Goebel <h.goebel@crazy-compilers.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import pprint
import tarfile
import shutil
import textwrap
import re

import warnings
warnings.filterwarnings('ignore', 'cannot parse package relationship .*, returning it raw')

from debian import deb822

from . import query
from .base32 import sha256

try:
    # "backport" xztar (was implemented in 3.5 only)
    import lzma
    del lzma
    shutil.register_unpack_format('xztar',['.tar.xz', '.txz'], 
                                  shutil._unpack_tarfile)
except ImportError:
    pass


def get_dependencies(ctlpkg, pkginfo):
    for dep in ('depends', 'build-depends', 'build-depends-indep'):
        if dep in ctlpkg:
            pkginfo[dep] = deb822.PkgRelation.parse_relations(ctlpkg[dep])
        else:
            pkginfo[dep] = []

def dsc2pkginfo(dsc):
    pkginfo = {
        'name': dsc['source'],
        'binaries': [b.strip() for b in dsc['binary'].split(',')],
        'version': dsc['version'],
        'homepage': dsc['homepage'],
    }
    files = [(f['name'], sha256(f['sha256']))
             for f in dsc['checksums-sha256']]
    pkginfo['patches'] = dict((n, c) for (n,c) in files if '.patch' in n)
    pkginfo['debianarch'] = dict((n, c) for (n,c) in files 
                                 if dsc['version'] + '.debian.tar' in n)
    pkginfo['sources'] = dict((n, c) for (n,c) in files 
                              if (n not in pkginfo['patches'] and 
                                  n not in pkginfo['debianarch']))
    assert len(pkginfo['debianarch']) == 1
    assert len(pkginfo['patches']) + len(pkginfo['debianarch']) + len(pkginfo['sources']) == len(files)

    get_dependencies(dsc, pkginfo)

    return pkginfo

def unpack_archive(archivefile, destdir):
    if not os.path.exists(destdir):
        os.makedirs(destdir)
    shutil.unpack_archive(archivefile, destdir)

WATCH_URL = re.compile(r'((http|https|ftp)://\S+)')

def get_watch_url(archive_dir, pkginfo):
    # collect relevant information from debian/watch file

    def get_watch(archive_dir):
        # :todo: adopt this to handle watchfile specification
        with open(os.path.join(archive_dir, 'debian', 'watch')) as fh:
            text = fh.readlines()
        # Parse accoding to `man uscan`
        last = None
        for l in text:
            l = l.lstrip() # Leading spaces and tabs are dropped
            if not l or l.startswith('#'):
                # Empty lines are dropped.
                # A line started by # (hash) is a comment line and dropped.
                continue
            if last is not None:
                l = last + l
                last = None
            if l.endswith('\\'):
                last = l
                continue
            m = WATCH_URL.search(l)
            if m:
                return m.group(0)

    url = get_watch(archive_dir)
    url = url.replace('\\.', '.')
    p = url.split('(.*)')
    pkginfo['uri'] = '"%s"' % '" version "'.join(p)


def get_packages(archive_dir, pkginfo):
    # collect relevant information from debian/control file
    with open(os.path.join(archive_dir, 'debian', 'control')) as fh:
        text = fh.readlines()
    pkginfo['package'] = pkgs = []
    for ctlpkg in deb822.Packages.iter_paragraphs(text):
        if 'source' in ctlpkg:
            continue
        pkg = {
            'name': ctlpkg['package'],
            'depends': ctlpkg['depends'],
            }
        pkgs.append(pkg)
        pkg['synopsis'], description = ctlpkg['description'].split('\n', 1)
        description = textwrap.dedent(description).strip()
        description = description.replace('\n.\n', '\n\n')
        pkg['description'] = description
        get_dependencies(ctlpkg, pkg)
    
def get_license(archive_dir, pkginfo):
    pkginfo['license'] = 'unknown'
    with open(os.path.join(archive_dir, 'debian', 'copyright')) as fh:
        text = fh.readlines()
    for l in text:
        if l.startswith('License: '):
            pkginfo['license'] = l.split(None, 1)[1].strip()

def collect_info(packagename, cachedir):
    vers = query.get_pkg_versions(packagename)
    vers.sort(reverse=True) # :todo: sort semantically
    ver, area = vers[0]
    dsc = query.get_pkg_dsc(packagename, ver, area)
    dsc = deb822.Dsc(dsc)

    pkginfo = dsc2pkginfo(dsc)

    # download and unpack the .debian archive
    filename = list(pkginfo['debianarch'].keys())[0]
    archivefile = os.path.join(cachedir, filename)
    query.store_pkg_debian_arch(filename, packagename, area, archivefile)
    archive_dir = os.path.join(cachedir, '{name}-{version}'.format(**pkginfo))
    unpack_archive(archivefile, archive_dir)

    get_watch_url(archive_dir, pkginfo)
    get_packages(archive_dir, pkginfo)
    get_license(archive_dir, pkginfo)

    return pkginfo, archive_dir

if __name__ == '__main__':
    packagename = 'commons-io'
    cachedir = '/tmp'
    pkginfo, outfilename = collect_info(packagename, cachedir)

    pprint.pprint(pkginfo)
    print(outfilename)



# -*- coding: utf-8 -*-
# Guix Import Debian --- Import helper for GNU Guix
# Copyright © 2016 Hartmut Goebel <h.goebel@crazy-compilers.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

__version__ = '0.1'

from . import deb
from .mappings import DEBIAN_LICENSE_MAP

import pprint

template = """
(define-public {name}-{version}
  (package
    (name "{name}")
    (version "{version}")
    (source (origin
      (method url-fetch)
      (uri (string-append
            {uri}))
      (sha256 (base32 "{sha256}"))))
    (build-system {build_system})
    (inputs
     `({inputs}))
    (native-inputs
     `({native_inputs}))
    (homepage "{homepage}")
    (synopsis "{synopsis}")
    (description "{description}")
    (license {license})
))
"""

class Input:
    def __init__(self, name, version=None, comment=None, **kw):
        self.name = name
        self.version = ' '.join(version or ())
        self.comment = comment

    def __eq__(self, other):
        return self.name.__eq__(other)

    def __str__(self):
        s = '("{0.name}" ,{0.name})'.format(self)
        if self.version and self.comment:
            s = s + ' ; {0.version} - {0.comment}'.format(self)
        elif self.version:
            s = s + ' ; {0.version}'.format(self)
        elif self.comment:
            s = s + ' ; {0.comment}'.format(self)
        return s

    def __repr__(self):
        if self.version:
            return 'Input(name={0.name!r}, version={0.version!r})'.format(self)
        else:
            return 'Input(name={0.name!r})'.format(self)

DEPENDS_TO_IGNORE = ['${misc:Depends}']

class Inputs(list):
    def __init__(self, pkginfo, *kind):

        def collect_inputs(pkg):
            for k in kind:
                for i in pkg[k]:
                    comment = None
                    if i[0]['name'] in names:
                        continue
                    names.add(i[0]['name'])
                    if len(i) > 1:
                        comment = ' | '. join(str(j for j in i))
                    inputs.append(Input(comment=comment, **i[0]))

        inputs = []
        names = set(DEPENDS_TO_IGNORE)

        collect_inputs(pkginfo)
        for pkg in pkginfo['package']:
            collect_inputs(pkg)

        super().__init__(inputs)

    def remove(self, input):
        '''Remove element if it exists.'''
        if isinstance(input, Input):
            input = input.name
        for i, ip in enumerate(self):
            if ip == input:
                del self[i]        

    def __str__(self):
        if self:
            return '\n\t'.join(str(i) for i in self) + '\n\t'
        else:
            return ''

LANGUAGES_TO_STRIP_FROM_NAME = ('python', 'java', 'ruby')

def guess_language(pkginfo):
    name = pkginfo['name']
    for lang in LANGUAGES_TO_STRIP_FROM_NAME:
        if name.startswith(lang + '-') or name == lang:
            # already in required format
            return None
        elif lang in name:
            return lang
    # use the first package's name to check if this is a library
    if not pkginfo['package'][0]['name'].startswith('lib'):
        return None
    if 'ant' in pkginfo['native_inputs']:
        return 'java'
    # did not find a language
    return None

def change_name_for_language(pkginfo):
    lang = guess_language(pkginfo)
    if lang:
        if lang in LANGUAGES_TO_STRIP_FROM_NAME:
            pkginfo['name'] = lang + '-' + pkginfo['name'].replace(lang, '')
        else:
            pkginfo['name'] = lang + '-' + pkginfo['name']

def guess_build_system(pkginfo):
    build_system = 'trivial-build-system'
    to_remove = []
    if 'ant' in pkginfo['native_inputs']:
        build_system = 'ant-build-system'
        to_remove= ['ant', 'default-jdk']
    elif pkginfo['name'].startswith('python-'):
        build_system = 'python-build-system'
        to_remove= ['python']
    elif pkginfo['name'].startswith('perl-'): # :fixme:
        build_system = 'perl-build-system'
        to_remove= ['perl']
    pkginfo['build_system'] = build_system
    for i in to_remove:
        pkginfo['native_inputs'].remove(i)


def pkginfo2package(packagename, cachedir):
    pkginfo, outfilename = deb.collect_info(packagename, cachedir)
    # assume the first source in the list is the "main" source
    pkginfo['sha256'] = pkginfo['sources'][list(pkginfo['sources'].keys())[0]]
    # use the first package's synsopsis and description
    pkginfo['synopsis'] = pkginfo['package'][0]['synopsis']
    pkginfo['description'] = pkginfo['package'][0]['description']
    pkginfo['version'], pkginfo['revision'] = pkginfo['version'].split('-',1)

    pkginfo['inputs'] = Inputs(pkginfo, 'depends')
    pkginfo['native_inputs'] = Inputs(pkginfo, 'build-depends', 'build-depends-indep')
    for i in pkginfo['inputs']:
        pkginfo['native_inputs'].remove(i)

    if pkginfo['license'] in DEBIAN_LICENSE_MAP:
        pkginfo['license'] = DEBIAN_LICENSE_MAP.get(pkginfo['license'])

    change_name_for_language(pkginfo)
    guess_build_system(pkginfo)

    print(template.format(**pkginfo))



# -*- coding: utf-8 -*-
# Guix Import Debian --- Import helper for GNU Guix
# Copyright © 2016 Hartmut Goebel <h.goebel@crazy-compilers.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

DEBIAN_LICENSE_MAP = {
    'Apache 2.0': 'asl2',
    'Apache-2': 'asl2',
}

DEBIAN_PACKAGE_MAP = {
    'default-jdk': ("jdk" ,'icedtea "jdk"'),
}



# -*- coding: utf-8 -*-
# Guix Import Debian --- Import helper for GNU Guix
# Copyright © 2016 Hartmut Goebel <h.goebel@crazy-compilers.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import json
import requests

suite = "jessie" # current debian release

DOWNLOAD_URL_BASE = 'http://ftp.uni-erlangen.de/debian/'

class NotFoundError(Exception): pass

def get_all_packages(suite=suite):
    "Returns a list of all package-names for this suite"
    raise NotImplementedError
    #'http://sources.debian.net/api/list/?suite=%s' % suite

def search_packages(query, suite=suite):
    "Return a list of package-names matching query"
    raise NotImplementedError
    # 'http://sources.debian.net/api/search/%s' % query


def get_pkg_versions(pkgname, suite=suite):
    """
    Return a list of (version, area) tuples.

    'Area' is e.g. 'main', 'non-free'
    """
    """
{u'package': u'ocaml',
 u'path': u'ocaml',
 u'pathl': [[u'ocaml', u'/src/ocaml/']],
 u'suite': u'jessie',
 u'type': u'package',
 u'versions': [{u'area': u'main',
                u'suites': [u'jessie'],
                u'version': u'4.01.0-5'}]}
"""
    r = requests.get("http://sources.debian.net/api/src/%s/?suite=%s" % 
                     (pkgname, suite))
    data = r.json()
    if 'error' in data:
        raise NotFoundError(data['error'])
    return [(v['version'], v['area']) for v in data['versions']]


def get_pkg_dsc(pkgname, version, area):
    #http://ftp.….de/debian/pool/non-free/a/arb/arb_6.0.3-1.dsc
    url = "{base}/pool/{area}/{name[0]}/{name}/{name}_{version}.dsc".format(
        base=DOWNLOAD_URL_BASE, name=pkgname, version=version, area=area)
    print(url)
    r = requests.get(url)
    return r.text


def store_pkg_debian_arch(filename, pkgname, area, outfilename):
    #http://ftp…de/debian/pool/non-free/a/arb/arb_6.0.2-1+deb8u1.debian.tar.xz
    url = "{base}/pool/{area}/{name[0]}/{name}/{filename}".format(
        base=DOWNLOAD_URL_BASE, name=pkgname, filename=filename, area=area)
    print(url)
    r = requests.get(url)
    if r.status_code == 200:
        with open(outfilename, 'wb') as f:
            for chunk in r.iter_content(10240):
                f.write(chunk)


if __name__ == '__main__':
    try:
        print(get_pkg_versions('commonsXX-io'))
    except Exception as e:
        print(repr(e))

    name = 'commons-io'
    v1 = get_pkg_versions(name)
    print(v1)
    dsc = get_pkg_dsc(name, *v1[0])
    print(dsc)
    filename = 'commons-io_2.4-2.debian.tar.gz'
    outfilename = os.path.join('/tmp', filename)
    store_pkg_debian_arch(filename, name, v1[0][1], outfilename)



# -*- coding: utf-8 -*-
# Guix Import Debian --- Import helper for GNU Guix
# Copyright © 2016 Hartmut Goebel <h.goebel@crazy-compilers.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# nix-type base32

import numconv

# See `libutil/hash.cc'.
nix_base32_chars = "0123456789abcdfghijklmnpqrsvwxyz"

_nix_base32 = numconv.NumConv(32, nix_base32_chars)

def nix_base32(int):
    return _nix_base32.int2str(int)

class sha256:
    def __init__(self, chksum):
        assert type(chksum) in (str, int)
        if type(chksum) is int:
            self.chksum = chksum
        elif len(chksum) == 64:
            # assume hex encoded
            self.chksum = int(chksum, 16)
        elif len(chksum) == 52:
            # assume base32 encoded
            # use nix encoding
            self.chksum = _nix_base32.str2int(chksum)

    def __str__(self):
        return self.base32()

    def __repr__(self):
        return "sha256('%s')" % self.base32()

    def base32(self):
        return nix_base32(self.chksum)

    def base16(self):
        # hex encoded
        return '%x' % self.chksum

# small tests
if __name__ == '__main__':
    h16 = 'a5e4625aef65ef7f23f54eabc1eda0748cbfa2b472976701a7e197bbe844497a'
    b32 = '19g4c9dfyrgggwizakmbq7ns0x4cpyib8wlpcw0sgqcppgl48jbs'
    c1 = sha256(h16)
    print(c1.base32())
    print(c1.base16())
    assert c1.base32() == b32
    assert c1.base16() == h16
    c2 = sha256(b32)
    assert c2.base32() == b32
    assert c2.base16() == h16



# -*- coding: utf-8 -*-
# Guix Import Debian --- Import helper for GNU Guix
# Copyright © 2016 Hartmut Goebel <h.goebel@crazy-compilers.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from setuptools import setup, find_packages

version = "0.1"

with open('README.rst') as fh:
    long_description = fh.read()

setup(
    name="guix-import-debian",
    version=version,
    description="""Helper for converting Debian packages into Guix package definitions""",
    long_description=long_description,
    author="Hartmut Goebel",
    author_email='h.goebel@crazy-compilers.com',
    license="GPLv3+",
    packages=find_packages(),
    include_package_data=True,
    install_requires=['python-debian', 'numconv', 'requests',
                      'chardet', # required by python-debian but not declared there
                  ],
    url="https://gitlab.com/htgoebel/guix-import-debian",
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Natural Language :: English',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python',
        ],
    entry_points = {
        "console_scripts": [
            "guix-import-debian = guix_import_debian.__main__:debian",
        ],
    },
    )
